using System;

namespace PropertyModifiers
{
    public class UsableObject : IUseable
    {
        public event EventHandler<UsageEventArgs> StartUsage;
        public event EventHandler<UsageEventArgs> StopUsage;

        private TimeSpan lastTime;
        private TimeSpan usageTimeSpan;
        private TimeSpan lastCooldownTime;
        private TimeSpan cooldownTimeSpan;
        bool inUse = false;
        private bool firstUpdate = false;
        protected IEventService events;
        protected TimeSpan usageDuration = TimeSpan.Zero;
        protected int usageLimit = int.MinValue;
        private bool isOffCooldown = true;
        private TimeSpan remainingTimeOnCooldown = TimeSpan.Zero;
        private readonly UsageEventArgs args = new UsageEventArgs();


        public virtual TimeSpan RemainingTimeOnCooldown
        {
            get { return remainingTimeOnCooldown; }
        }

        public virtual TimeSpan Cooldown
        {
            get { return this.cooldownTimeSpan; }
        }

        /// <summary>
        /// Returns the number of charges left for the item.
        /// </summary>
        /// <returns>Returns the number of remaining charges. If the valus is int.MinValue then there is no limit on it's use.</returns>
        public virtual int Charges
        {
            get { return this.usageLimit == int.MinValue ? int.MinValue : this.usageLimit; }
        }

        public virtual void SetCooldown(TimeSpan value)
        {
            this.cooldownTimeSpan = value;
            if (this.isOffCooldown) this.remainingTimeOnCooldown = this.cooldownTimeSpan;
        }

        public virtual void SetDuration(TimeSpan value)
        {
            this.usageDuration = value;
        }

        public virtual void SetCharges(int value)
        {
            this.usageLimit = value;
        }

        public virtual void Start()
        {
            if (this.inUse || this.isOffCooldown == false || this.usageLimit == 0)
            {
                this.events.Report(Category.Other, MessageType.Information, "Item cannot be used right now.");
                return;
            }

            this.inUse = true;
            if (this.cooldownTimeSpan > TimeSpan.Zero) this.isOffCooldown = false;
            this.args.ReturnValue = false;
            if (this.StartUsage != null) this.StartUsage(this, this.args);
            if (this.args.ReturnValue)
            {
                if (this.usageLimit != int.MinValue && this.usageLimit > 0)
                {
                    this.usageLimit--;
                }
                this.firstUpdate = true;
                this.usageTimeSpan = TimeSpan.Zero;
                this.events.ItemUpdate += this.Update;
            }
            else
            {
                this.inUse = false;
            }
        }

        public virtual void Stop()
        {
            this.inUse = false;
            this.args.ReturnValue = true;
            if (this.StopUsage != null) this.StopUsage(this, this.args);
            // if still on cooldown don't unhook event yet
            if (this.isOffCooldown ) this.events.ItemUpdate -= this.Update;
        }

        private void Update(object sender, ItemEventArgs e)
        {
            if (this.inUse)
            {
                if (this.firstUpdate)
                {
                    this.firstUpdate = false;
                    this.lastTime = e.GameTime;
                    this.lastCooldownTime = e.gameTime;
                }

                this.usageTimeSpan += e.GameTime - lastTime;
                lastTime = e.GameTime;
                this.remainingTimeOnCooldown = (this.lastCooldownTime + this.cooldownTimeSpan) - e.gameTime;

                if (this.usageTimeSpan >= this.usageDuration)
                {
                    this.inUse = false;
                    this.args.ReturnValue = true;
                    if (this.StopUsage != null) this.StopUsage(this, this.args);
                }
            }

            this.remainingTimeOnCooldown = (this.lastCooldownTime + this.cooldownTimeSpan) - e.gameTime;

            if (this.isOffCooldown == false && this.remainingTimeOnCooldown < TimeSpan.Zero)
            {
                this.isOffCooldown = true;
                this.remainingTimeOnCooldown = TimeSpan.Zero;
                this.events.ItemUpdate -= this.Update; 
            }
        }

        public UsableObject(IGameService service)
        {
            this.events = service.Events;
        }

        public virtual string GetDescription()
        {
            int charges = this.Charges;
            string chargeString = charges >= 0 ? " ( Charges: " + charges + " )" : string.Empty;

            string cooldownString = this.cooldownTimeSpan > TimeSpan.Zero ? " ( " + Helpers.TimeSpanToFriendlyString(this.Cooldown) + " cooldown )" : string.Empty;
            if (this.isOffCooldown == false) cooldownString = " ( " + Helpers.TimeSpanToFriendlyString(this.RemainingTimeOnCooldown) + " remaining )";

            return chargeString + cooldownString;
        }
    }
}